;
; Music player for NES
; Copyright 2011 Damian Yerrick
;
; Copying and distribution of this file, with or without
; modification, are permitted in any medium without royalty provided
; the copyright notice and this notice are preserved in all source
; code copies.  This file is offered as-is, without any warranty.
;

.include "src/nes.h"
.include "src/ram.h"

NUM_SONGS = 2

.import cart_menu, get_titledir_a
.import clearLineImg, copyLineImg, vwfPuts
.import interbank_fetch, interbank_fetch_buf
.importzp lineImgBufLen
.exportzp psg_sfx_state

; ld65 manual says LOWCODE is for stuff guaranteed not to be
; banked out, good for irq handlers, but it's also good for
; anything that must access data from multiple banks
.import __LOWCODE_RUN__, __LOWCODE_LOAD__
.import __LOWCODE_SIZE__

.segment "ZEROPAGE"
nmis:          .res 1
oam_used:      .res 1

; Used by pads.s
cur_keys:  .res 2
new_keys:  .res 2
cur_trigger: .res 1

; Used by music engine
psg_sfx_state: .res 32
tvSystem:   .res 1

.segment "BSS"
linebuf: .res 40
ciBlocksLeft = linebuf

.segment "INESHDR"
  .byt "NES",$1A  ; magic signature
  .byt 2          ; PRG ROM size in 16384 byte units
  .byt 0          ; CHR ROM size in 8192 byte units
  .byt $00        ; mirroring type and mapper number lower nibble
  .byt $00        ; mapper number upper nibble

.segment "BFF0"
.proc patch_bff0
  sei
ldxinstr:
  ldx #$FF
  txs
  stx ldxinstr+1
  jmp ($FFFC)
.endproc

.segment "FFF0"
.proc patch_fff0
  sei
ldxinstr:
  ldx #$FF
  txs
  stx ldxinstr+1
  jmp ($FFFC)
  .addr nmi, reset, irq
.endproc

; Add this only when copying hex into the patcher
.if 1
.segment "CODE"
.proc patch21
  sei
  ldx #copied_end - copied - 1
:
  lda copied,x
  sta $00,x
  dex
  bpl :-
  jmp $0000
copied:
  stx copied+5
  jmp ($FFFC)
copied_end:
.endproc
.endif

.segment "CODE"
;;
; This NMI handler is good enough for a simple "has NMI occurred?"
; vblank-detect loop, like in Lawn Mower or Thwaite.
.proc nmi
  inc nmis
  rti
.endproc

; A null IRQ handler that just does RTI is useful to add breakpoints
; that survive a recompile: trap reads of $FFFE, and then BRK $00.
; But sometimes you'll want a non-null IRQ handler, be it for the
; DMC IRQ or a mapper IRQ.
.proc irq
  rti
.endproc

; 
.proc reset
  ; The very first thing to do when powering on is to put all sources
  ; of interrupts into a known state.
  sei             ; Disable interrupts
  ldx #$00
  stx PPUCTRL     ; Disable NMI and set VRAM increment to 32
  stx PPUMASK     ; Disable rendering
  stx $4010       ; Disable DMC IRQ
  dex             ; Subtracting 1 from $00 gives $FF, which is a
  txs             ; quick way to set the stack pointer to $01FF
  bit PPUSTATUS   ; Acknowledge stray vblank NMI across reset
  bit SNDCHN      ; Acknowledge DMC IRQ
  lda #$40
  sta P2          ; Disable APU Frame IRQ
  lda #$0F
  sta SNDCHN      ; Disable DMC playback, initialize other channels

vwait1:
  bit PPUSTATUS   ; It takes one full frame for the PPU to become
  bpl vwait1      ; stable.  Wait for the first frame's vblank.

  ; Turn off decimal mode for post-patent famiclones
  cld
  jsr init_sound

vwait2:
  bit PPUSTATUS  ; After the second vblank, we know the PPU has
  bpl vwait2     ; fully stabilized.  After this use only NMI.

  lda #VBLANK_NMI
  sta PPUCTRL
  jsr getTVSystem
  sta tvSystem
  
  ; Copy the CHR decompression code to RAM.  This is limited to
  ; 256 bytes, and I don't know how to make a link-time assertion
  ; to ensure this.
  ldx #0
copy_LOWCODE:
  lda __LOWCODE_LOAD__,x
  sta __LOWCODE_RUN__,x
  inx
  cpx #<__LOWCODE_SIZE__
  bne copy_LOWCODE
  
  lda #0
  jsr init_music
  jsr title_screen
  jsr cart_menu
  ldx #0
  stx SNDCHN
  ; game id is in A
  pha
  jsr get_titledir_a
  jsr load_titledir_chr_rom
  pla
  jsr get_titledir_a
  jmp jump_to_titledir
.endproc

line1dst = $0E08
line2dst = line1dst + 2*lineImgBufLen
.segment "CODE"
.proc title_screen
  lda PPUSTATUS
  lda #VBLANK_NMI
  sta PPUCTRL

  ; fill in the palette and first nametable
  ldy #$00
  tya
  sta PPUMASK
  ldx #$3F
  stx PPUADDR
  sta PPUADDR
  stx PPUDATA
  sta PPUDATA
  ldx #$10
  stx PPUDATA
  ldx #$20
  stx PPUDATA
  jsr ppu_clear_nt

  lda #$00
  sta PPUADDR
  sta PPUADDR
  ldy #0
ol:
.if 0
  ldx #15
  stx PPUDATA
  tya
.else
  ldx #16
.endif
il:
  sta PPUDATA
  dex
  bne il
  iny
  bne ol


  ; Make spaces into which to draw
  ; "Press Start" and the author notice

  ldy #<(line1dst >> 4)
  lda #$22
  sta PPUADDR
  lda #$88
  sta PPUADDR
  ldx #<(lineImgBufLen/8)
mknt1:
  sty PPUDATA
  iny
  dex
  bne mknt1

  lda #$23
  sta PPUADDR
  lda #$08
  sta PPUADDR
  ldx #lineImgBufLen/8
mknt2:
  sty PPUDATA
  iny
  dex
  bne mknt2

  lda #$05
  sta 0
  ldx #0
logorowloop:
  ldy #22
  lda #$21
  sta PPUADDR
  lda 0
  sta PPUADDR
  clc
  adc #32
  sta 0
logotileloop:  
  lda a53logo,x
  sta PPUDATA
  inx
  dey
  bne logotileloop
  cpx #84
  bcc logorowloop

  ; Load non-text tiles used for title screen
  ldy #$00
  sty PPUADDR
  sty PPUADDR
  lda #<logotiles_chr
  sta 0
  lda #>logotiles_chr
  sta 1
  ldx #2
copyseltilesloop:
  lda (0),y
  sta PPUDATA
  iny
  bne copyseltilesloop
  inc 1
  dex
  bne copyseltilesloop
  
  jsr clearLineImg
  lda #>hello_msg
  ldy #<hello_msg
  ldx #20
  jsr vwfPuts
  lda #>line1dst
  ldy #<line1dst
  jsr copyLineImg

  jsr clearLineImg
  lda #>hello2_msg
  ldy #<hello2_msg
  ldx #0
  jsr vwfPuts
  lda #>line2dst
  ldy #<line2dst
  jsr copyLineImg

title_wait_A:
  lda cur_keys
  asl a
  and #BG_1000
  ora #VBLANK_NMI
  sta PPUCTRL
  jsr ppu_wait_vblank
  lda #$00
  sta PPUSCROLL
  sta PPUSCROLL
  lda #BG_ON
  sta PPUMASK
  jsr read_pads
  ldx #0
  jsr read_mouse_with_backward_buttons
  jsr update_sound
  jsr read_zapper_trigger
  ora new_keys
  and #KEY_START|KEY_A
  beq title_wait_A
  rts
.endproc

.segment "RODATA"
hello_msg:  .byt "Press Start Button",0
hello2_msg: .byt "Menu 0.02 by D. Yerrick 2012",0

a53logo:
  .byt  0, 0, 0, 0, 0, 0,15, 0, 0, 0, 0, 0, 0, 0, 0,22,23,24,25,26,27, 0
  .byt  2, 3, 4, 2, 3, 4,16,17,20, 2, 3, 4, 2, 3, 4,28, 3, 4,29,30,31, 0
  .byt  5, 0,11, 5, 0, 0,18, 0,21, 5, 0, 6,10, 0,11, 0, 0, 6, 0, 0, 6, 0
  .byt  7,14,13, 7, 8, 9, 7,19,13, 7, 8, 9,12, 0,13, 7, 8, 9, 7, 8, 9, 1

logotiles_chr: .incbin "obj/nes/logotiles.chr"

.importzp ciSrc, ciDst, ciBufStart, ciBufEnd
.importzp ciSrc0, ciSrc1
.import draw_progress
.import unpb53_some, PB53_outbuf, interbank_fetch, interbank_fetch_buf

.segment "CODE"
CHRDIR_START = $FF02
; each entry is bank, address (16-bit), midpoint offset (16-bit)
.proc load_titledir_chr_rom
pointer_to_bank_number = interbank_fetch_buf+72
  lda #0
  sta draw_progress
  sta PPUMASK
  sta PPUCTRL
  ldy #1
  lda (0),y  ; get CHR bank id into A
  bpl is_chr_rom  ; $80-$FF means absent
  rts
is_chr_rom:
  ldy #0
  sty 3
  sta 2
  asl a
  asl a
  rol 3
  adc 2
  bcc :+
  inc 3
:
  clc
  adc CHRDIR_START
  sta 2
  sta pointer_to_bank_number
  lda CHRDIR_START+1
  adc 3
  sta 3
  sta pointer_to_bank_number+1

  ldy #1
  lda (2),y
  iny
  sta ciSrc0
  lda (2),y
  iny
  sta ciSrc0+1
  clc
  lda (2),y
  iny
  adc ciSrc0
  sta ciSrc1
  lda (2),y
  adc ciSrc0+1
  sta ciSrc1+1

  ; At this point:
  ; 2: pointer to bank number
  ; ciSrc0: Address of first 4 KiB
  ; ciSrc1: Address of second 4 KiB
  ; drawProgress: number of blocks already drawn
loop:
  lda ciSrc0
  sta 0
  lda ciSrc0+1
  sta 1
  lda #0
  jsr do4
  adc ciSrc0
  sta ciSrc0
  bcc :+
  inc ciSrc0+1
:

  lda ciSrc1
  sta 0
  lda ciSrc1+1
  sta 1
  lda #64
  jsr do4
  adc ciSrc1
  sta ciSrc1
  bcc :+
  inc ciSrc1+1
:
  lda #0
  sta 0
  lda draw_progress
  lsr a
  ror 0
  lsr a
  ror 0
  sta PPUADDR
  ora #$10
  sta 1
  lda 0
  sta PPUADDR
  ldx #0
copyloop0:
  lda PB53_outbuf,x
  sta PPUDATA
  inx
  cpx #64
  bcc copyloop0
  lda 1
  sta PPUADDR
  lda 0
  sta PPUADDR
copyloop1:
  lda PB53_outbuf,x
  sta PPUDATA
  inx
  cpx #128
  bcc copyloop1
  inc draw_progress
  lda draw_progress
  cmp #64
  bcc loop
  rts

do4:
  sta ciBufStart
  clc
  adc #64
  sta ciBufEnd
  lda #72  ; 4 tiles, max 18 bytes per tile
  sta 4
  lda pointer_to_bank_number
  sta 2
  lda pointer_to_bank_number+1
  sta 3
  jsr interbank_fetch
  lda #<interbank_fetch_buf
  sta ciSrc
  lda #>interbank_fetch_buf
  sta ciSrc+1
  jsr unpb53_some
  lda ciSrc
  sec
  sbc #<interbank_fetch_buf
  clc
  rts
.endproc

;;
; Reads a mouse and maps presses of L to A and R to B
; in new_keys.
; @param X port number of mouse to read (0 or 1)
; @return A zero iff mouse is present
.proc read_mouse_with_backward_buttons
  jsr read_mouse
  lda 1
  eor #$01
  and #$0F
  bne no_mouse

  lda new_mbuttons,x
  lsr a
  and #KEY_B
  ora new_keys,x
  sta new_keys,x
  lda new_mbuttons,x
  asl a
  and #KEY_A
  ora new_keys,x
  sta new_keys,x

  lda #0
no_mouse:
  rts
.endproc

.proc read_zapper_trigger
  lda $4017
  and #$10
  sta 0
  lda cur_trigger
  eor #$10
  and 0
  ldy 0
  sty cur_trigger
  rts
.endproc

.segment "LOWCODE"
.proc jump_to_titledir
  ldy #12
  lda (0),y
  sta 2
  iny
  lda (0),y
  sta 3
  ldy #0
  lda (0),y
  sta (0),y
  jmp ($0002)
.endproc

